@@ -33,13 +33,17 @@ __all__ = [
     # requirements
     'BaseRequirement',
     'Anything',
+    'Exists',
     'IsA',
+    'HasAttr',
     'Equals',
+    'Contains',
     'InRange',
     'Length',
     'ListOf',
+    'ListOfAll',
+    'ListOfAny',
     'DictOf',
-    'NotExists',
 
     # functions
     'translate',
@@ -49,24 +53,61 @@ __all__ = [
 ]
 
 
+import copy
+
 from . import compat
 from .errors import (
     CombinedValidationError, AtLeastOneFailed, AllFailed, ValidationError,
-    NoDefaultValue, InvalidKey, MissingKey, StructureSpecificationError
+    NoDefaultValue, InvalidKeys, MissingKeys, StructureSpecificationError,
+    DictValueError,
 )
 
 
+#: The value is valid if any of its items passes validation.
+ITEM_STRATEGY_ANY = 'any'
+#: The value is valid if all of its items pass validation.
+ITEM_STRATEGY_ALL = 'all'
+
+
+class MISSING:
+    """
+    Stub for Exists validator to pass if the value is missing
+    (e.g. for dictionary keys).
+    """
+    pass
+
+
+def _reluctantly_translate(spec):
+    # `translate()` can do it itself but some validators have the `implies`
+    # attribute which can trigger instantiation of a BaseValidator subclass
+    # before the translation function is ready.
+    #
+    # We don't want to defer its usage as far as we can because it is best
+    # to fully build the validator in order to fail early.
+    #
+    # So this function is just a small barrier that prevents NameError
+    # in some cases.
+
+    if isinstance(spec, BaseValidator):
+        return spec
+    else:
+        return translate(spec)
+
+
 class BaseValidator(object):
+    error_class = ValidationError
     _default = NotImplemented
+    negated = False
 
     def _combine(self, other, combinator):
         # XXX should we flatten same-logic one-item combs?
         if isinstance(other, type) and issubclass(other, BaseValidator):
-            # e.g. NotExists instead of NotExists()
+            # e.g. Exists instead of Exists()
             raise TypeError('got {cls} class instead of its instance'
                             .format(cls=other.__name__))
 
-        return combinator([self, translate(other)])
+
+        return combinator([self, _reluctantly_translate(other)])
 
     def _merge(self, value):
         if value is not None:
@@ -86,6 +127,23 @@ class BaseValidator(object):
     def __eq__(self, other):
         return isinstance(other, type(self)) and self.__dict__ == other.__dict__
 
+    def __invert__(self):
+        clone = copy.deepcopy(self)
+        clone.negated = not self.negated
+        return clone
+
+    def __call__(self, value):
+        try:
+            self._check(value)
+        except ValidationError:
+            if self.negated:
+                return
+            else:
+                raise
+        else:
+            if self.negated:
+                self._raise_error(value)
+
     def __hash__(self):
         # TODO think this over and check Python docs
         #return hash(((k,v) for k,v in self.__dict__.items()))
@@ -100,19 +158,26 @@ class BaseValidator(object):
             else:
                 raise
 
+    def _check(self, value):
+        raise NotImplementedError
+
+    def _raise_error(self, value):
+        raise self.error_class(repr(self))
+
 
 class BaseCombinator(BaseValidator):
     error_class = CombinedValidationError
     break_on_first_fail = False
+    _repr_tmpl = '{not_}({items})'
+    _repr_items_sep = '; '
 
     def __init__(self, specs, default=None, first_is_default=False):
         assert specs
-
-        self._specs = [translate(s) for s in specs]
+        self._specs = [_reluctantly_translate(s) for s in specs]
         self._default = default
         self._first_is_default = first_is_default
 
-    def __call__(self, value):
+    def _check(self, value):
         errors = []
         for spec in self._specs:
             # TODO: group errors by exception type
@@ -131,23 +196,14 @@ class BaseCombinator(BaseValidator):
                     raise
                 errors.append(e)
         if not self.can_tolerate(errors):
-            def fmt_err(e):
-                # only display error class if it's not obvious
-                if type(e) is ValidationError:
-                    tmpl = '{err}'
-                else:
-                    tmpl = '{cls}: {err}'
-                return tmpl.format(cls=e.__class__.__name__, err=e)
-
-            errors_str = '; '.join((fmt_err(e) for e in errors))
-            raise self.error_class(
-                '{value!r} ({errors})'.format(value=value, errors=errors_str))
+            raise self.error_class(*errors)
 
 
     def __repr__(self):
-        return '{cls}[{validators}]'.format(
+        return self._repr_tmpl.format(
             cls=self.__class__.__name__,
-            validators=', '.join(map(str, self._specs)))
+            items=self._repr_items_sep.join(map(str, self._specs)),
+            not_='not ' if self.negated else '')
 
 
     def can_tolerate(self, errors):
@@ -181,6 +237,7 @@ class All(BaseCombinator):
     """
     error_class = AtLeastOneFailed
     break_on_first_fail = True
+    _repr_items_sep = ' and '
 
     def can_tolerate(self, errors):
         # TODO: fail early, work as `or` does
@@ -194,20 +251,13 @@ class Any(BaseCombinator):
     Requires that the value passes at least one of nested validators.
     """
     error_class = AllFailed
+    _repr_items_sep = ' or '
 
     def can_tolerate(self, errors):
         if len(errors) < len(self._specs):
             return True
 
 
-class MISSING:
-    """
-    Stub for NotExists validator to pass if the value is missing
-    (e.g. for dictionary keys).
-    """
-    pass
-
-
 class BaseRequirement(BaseValidator):
     # a hint for combinators, see their code
     is_recursive = False
@@ -216,15 +266,16 @@ class BaseRequirement(BaseValidator):
     def __call__(self, value):
         if self.implies is not NotImplemented:
             self.implies(value)
-        self._check(value)
+        super(BaseRequirement, self).__call__(value)
 
     def _represent(self):
         return self.__dict__
 
     def __repr__(self):
-        return '{cls}({rep})'.format(cls=self.__class__.__name__,
-                                     rep=self._represent())
-
+        return '{negated}{cls}({rep})'.format(
+            cls=self.__class__.__name__,
+            rep=self._represent(),
+            negated='~' if self.negated else '')
 
 
 class Anything(BaseRequirement):
@@ -248,16 +299,13 @@ class IsA(BaseRequirement):
 
     def _check(self, value):
         if not isinstance(value, self.expected_type):
-            raise ValidationError('must be {type}'
-                                  .format(type=self.expected_type.__name__))
+            self._raise_error(value)
 
-    def _represent(self):
-        if self._default:
-            return '{name} {default!r}'.format(
-                name=self.expected_type.__name__,
-                default=self._default)
-        else:
-            return self.expected_type.__name__
+    def __repr__(self):
+        s = 'must be {pattern_}'.format(pattern_=self.expected_type.__name__)
+        if self.negated:
+            s = 'not ({s})'.format(s=s)
+        return s
 
 
 class Equals(BaseRequirement):
@@ -269,20 +317,44 @@ class Equals(BaseRequirement):
 
     def _check(self, value):
         if self._expected_value != value:
-            raise ValidationError('!= {expected!r}'
-                                  .format(expected=self._expected_value))
+            self._raise_error(value)
 
-    def _represent(self):
-        return repr(self._expected_value)
+    def __repr__(self):
+        s = 'must equal {pattern_!r}'.format(pattern_=self._expected_value)
+        if self.negated:
+            s = 'not ({s})'.format(s=s)
+        return s
 
     @property
     def _default(self):
         return self._expected_value
 
 
-class NotExists(BaseRequirement):
+class Contains(BaseRequirement):
     """
-    Requires that the value does not exist.  Obviously this only makes sense in
+    Requires that the value contains given expected value.
+    """
+    def __init__(self, expected_value):
+        self._expected_value = expected_value
+
+    def _check(self, value):
+        if self._expected_value not in value:
+            self._raise_error(value)
+
+    def __repr__(self):
+        s = 'must contain {pattern_!r}'.format(pattern_=self._expected_value)
+        if self.negated:
+            s = 'not ({s})'.format(s=s)
+        return s
+
+    @property
+    def _default(self):
+        return self._expected_value
+
+
+class Exists(BaseRequirement):
+    """
+    Requires that the value exists.  Obviously this only makes sense in
     special cases like dictionary keys; otherwise there's simply nothing to
     validate.  Note that this is *not* a check against `None` or `False`.
     """
@@ -290,30 +362,26 @@ class NotExists(BaseRequirement):
         self._default = default
 
     def _check(self, value):
-        if value is not MISSING:
-            raise ValidationError('must not exist')
+        if value is MISSING:
+            self._raise_error(value)
 
-    def _represent(self):
-        return ''
+    def __repr__(self):
+        if self.negated:
+            return 'must not exist'
+        else:
+            return 'must exist'
 
 
-
-class ListOf(BaseRequirement):
+class BaseListOf(BaseRequirement):
     """
-    Requires that the value is a `list` which items match given validator.
-    Usage::
-
-        >>> v = ListOf(IsA(int) | IsA(str))
-        >>> v([123, 'hello'])
-        >>> v([123, 'hello', 5.5])
-        Traceback (most recent call last):
-        ...
-        ValidationError: #2: 5.5 (ValidationError: must be int;
-                                  ValidationError: must be str)
-
+    The base class for validating lists.  Supports different error toleration
+    strategies which can be selected by subclasses.  In many aspects this is
+    similar to :class:`BaseCombinator`.
     """
-    is_recursive = True
     implies = IsA(list)
+    item_strategy = NotImplemented
+    error_class = CombinedValidationError
+    is_recursive = True
 
     def __init__(self, validator, default=None):
         self._nested_validator = translate(validator)
@@ -324,14 +392,38 @@ class ListOf(BaseRequirement):
             try:
                 self._nested_validator(MISSING)
             except ValidationError as e:
-                raise ValidationError('missing element: {error}'
+                raise ValidationError('lacks item: {error}'
                                       .format(error=e))
+
+        errors = []
         for i, nested_value in enumerate(value):
             try:
                 self._nested_validator(nested_value)
             except ValidationError as e:
-                raise ValidationError('#{elem}: {error}'
-                                      .format(elem=i, error=e))
+                annotated_error = ValidationError(
+                    'item #{elem}: {error}'.format(elem=i, error=e))
+                if self.item_strategy == ITEM_STRATEGY_ALL:
+                    raise annotated_error
+                errors.append(annotated_error)
+
+        if self.can_tolerate(errors, value):
+            return
+
+        raise self.error_class(*errors)
+
+    def can_tolerate(self, errors, value):
+        if self.item_strategy == ITEM_STRATEGY_ALL:
+            if errors:
+                return False
+            else:
+                return True
+        elif self.item_strategy == ITEM_STRATEGY_ANY:
+            if len(errors) < len(value):
+                return True
+            else:
+                return False
+        else:
+            raise ValueError('unknown strategy')
 
     def _represent(self):
         return repr(self._nested_validator)
@@ -355,6 +447,35 @@ class ListOf(BaseRequirement):
         return [x if x is None else item_spec.get_default_for(x) for x in value]
 
 
+class ListOfAll(BaseListOf):
+    """
+    Requires that the value is a `list` which items match given validator.
+    Usage::
+
+        >>> v = ListOfAll(IsA(int) | IsA(str))
+        >>> v([123, 'hello'])
+        >>> v([123, 'hello', 5.5])
+        Traceback (most recent call last):
+        ...
+        ValidationError: #2: 5.5 (ValidationError: must be int;
+                                  ValidationError: must be str)
+
+    """
+    error_class = AtLeastOneFailed
+    item_strategy = ITEM_STRATEGY_ALL
+
+
+class ListOfAny(BaseListOf):
+    """
+    Same as :class:`ListOfAll` but tolerates invalid items as long as there
+    is at least one valid among them.
+    """
+    error_class = AllFailed
+    item_strategy = ITEM_STRATEGY_ANY
+
+
+ListOf = ListOfAll
+
 
 #@requirement(implies=[IsA(dict)], is_recursive=True, vars=['key', 'req'])
 #def dict_contains(ctx, value):
@@ -362,8 +483,6 @@ class ListOf(BaseRequirement):
 #    ctx['req'](nested_value)
 
 
-
-
 class DictOf(BaseRequirement):
     """
     Requires that the value is a `dict` which items match given patterns.
@@ -373,7 +492,7 @@ class DictOf(BaseRequirement):
         ...     # key "name" must exist; its value must be a `str`
         ...     (Equals('name'), IsA(str)),
         ...     # key "age" may not exist; its value must be an `int`
-        ...     (Equals('age') | NotExists(), IsA(int)),
+        ...     (Equals('age') | ~Exists(), IsA(int)),
         ...     # there may be other `str` keys with `str` or `int` values
         ...     (IsA(str), IsA(str) | IsA(int)),
         ... ])
@@ -390,7 +509,7 @@ class DictOf(BaseRequirement):
         AllFailed: 'note': 5.5 (ValidationError: must be str;
                                 ValidationError: must be int)
 
-    Note that this validator supports :class:`NotExists` to mark keys that can
+    Note that this validator supports :class:`Exists` to mark keys that can
     be missing.
     """
     implies = IsA(dict)
@@ -416,19 +535,24 @@ class DictOf(BaseRequirement):
                 if k in validated_data_keys:
                     continue
 
-                # check if this key is described by current rule;
-                # if it isn't, just skip it (and try another rule on it later on)
+                # check if this key is described by current key validator;
+                # if it isn't, just skip it (and try another validator
+                # on it later on)
                 try:
                     k_validator(k)
                 except (TypeError, ValidationError):
                     continue
 
-                # this key *is* described by current rule;
+                # this key *is* described by current value validator;
                 # validate the value (it *must* validate)
                 try:
                     v_validator(v)
                 except (ValidationError, TypeError) as e:
-                    raise type(e)('{k!r}: {e}'.format(k=k, e=e))
+                    if isinstance(e, DictValueError):
+                        msg = 'in {k!r} ({e})'
+                    else:
+                        msg = '{k!r} value {e}'
+                    raise DictValueError(msg.format(k=k, e=e))
 
                 validated_data_keys.append(k)
                 matched = True
@@ -445,12 +569,16 @@ class DictOf(BaseRequirement):
         # check if there are data keys that did not match any key spec;
         # if yes, raise InvalidKey for them
         if len(validated_data_keys) < len(value):
-            raise InvalidKey(', '.join(repr(x) for x in set(value) - set(validated_data_keys)))
+            invalid_keys = set(value) - set(validated_data_keys)
+            raise InvalidKeys(*invalid_keys)
 
         if missing_key_specs:
-            # NOTE: this prints rules, not keys as strings
-            raise MissingKey('{0}'.format(
-                ', '.join(compat.safe_str(rule) for rule in missing_key_specs)))
+            # XXX this prints validators, not keys as strings;
+            #     one exception is the Equals validator from which we get
+            #     the expected value via internal API.  And that's gross.
+            reprs = (spec._expected_value if isinstance(spec, Equals) else spec
+                     for spec in missing_key_specs)
+            raise MissingKeys(*reprs)
 
 
     def _merge(self, value):
@@ -493,37 +621,74 @@ class InRange(BaseRequirement):
     """
     Requires that the numeric value is in given boundaries.
     """
+    implies = IsA(int) | IsA(float)
+
     def __init__(self, min=None, max=None, default=NotImplemented):
         self._min = min
         self._max = max
-        self._default = default
+        if default is not NotImplemented:
+            self._default = default
 
     def _check(self, value):
-        if value is MISSING:
-            raise InvalidKey(value)
         if self._min is not None and self._min > value:
-            raise ValidationError('must be ≥ {expected}'
-                                  .format(expected=self._min))
+            self._raise_error(value)
         if self._max is not None and self._max < value:
-            raise ValidationError('must be ≤ {expected}'
-                                  .format(expected=self._max))
+            self._raise_error(value)
 
-    def _represent(self):
+    def __repr__(self):
+        if self.negated:
+            must = 'must not'
+        else:
+            must = 'must'
         def _fmt(x):
             return '' if x is None else x
-        return '{min}..{max}'.format(min=_fmt(self._min),
-                                     max=_fmt(self._max))
+        return '{must} belong to {min_}..{max_}'.format(
+            must=must, min_=_fmt(self._min), max_=_fmt(self._max))
+
+
+
+class HasAttr(BaseRequirement):
+    """
+    Requires that the value has given attribute.
+    """
+    def __init__(self, attr_name):
+        self._attr_name = attr_name
+
+    def _check(self, value):
+        if not hasattr(value, self._attr_name):
+            self._raise_error(value)
+
+    def __repr__(self):
+        if self.negated:
+            must = 'must not'
+        else:
+            must = 'must'
+        return '{must} have attribute {name!r}'.format(
+            must=must, name=self._attr_name)
+
 
 
 class Length(InRange):
     """
     Requires that the value length is in given boundaries.
     """
+    implies = HasAttr('__len__')
+
     def _check(self, value):
         try:
             super(Length, self)._check(len(value))
         except ValidationError as e:
-            raise ValidationError('length ' + str(e))
+            self._raise_error(value)
+
+    def __repr__(self):
+        if self.negated:
+            must = 'must not'
+        else:
+            must = 'must'
+        def _fmt(x):
+            return '' if x is None else x
+        return '{must} have length of {min_}..{max_}'.format(
+            must=must, min_=_fmt(self._min), max_=_fmt(self._max))
 
 
 def translate(value):
